#!/usr/bin/env python

""" A Class to handle MOM4 outputs
"""

from pydap.client import open_url

class Model(object):

    def __init__(self, urlpath):
        self.data = open_url(urlpath)
        self._field = None

    @property
    def field(self):
        return self._field

    @field.setter
    def field(self, field):
        self._field = self.data[field]

    @property
    def lon(self):
        LON = self.field.dimensions[-1]
        return self.data[LON][:]

    @property
    def lat(self):
        LAT = self.field.dimensions[-2]
        return self.data[LAT][:]

    @property
    def depth(self):
        DEPTH = self.field.dimensions[-3]
        return self.data[DEPTH][:]

    @property
    def time(self):
        TIME = self.field.dimensions[-4]
        return [(v, self.data[TIME].units) for v in self.data[TIME][:]]


class MOM4p1(Model):

    def __init__(self, urlpath):
        print urlpath
        super(MOM4p1, self).__init__(urlpath)
        self.name = 'MOM4p1'

    def get_data(self, l0, j0, i0):
        return ma.masked_equal(
                self.field.array[l0:l0 + 2, :, j0:j0 + 2, i0:i0 + 2],
                self.field.missing_value)

from UserDict import UserDict
import numpy
from numpy import ma

class Modelo_from_dap(UserDict):
    """
    """
    def __init__(self,urlpath):
        """
        """
        from pydap.client import open_url
        self.data = {}

        self.dataset = open_url(urlpath)
        ## Dictionary with the reference data for the model
        ##print "Available: ",dataset_model.keys()
        #for v in ['time','xt_ocean','yt_ocean','st_ocean']:
        #    self.data[v] = ma.array(self.dataset[v][:])
        ## Adjust the grid to the regular [-180:180]
        #self.data['xt_ocean'][self.data['xt_ocean']<-180] = self.da
        return

    def keys(self):
        keys = self.data.keys()
        for k in self.dataset.keys():
	    if k not in keys:
	        keys.append(k)
        return keys

    def __getitem__(self,index):
        """
        """
	if index in self.data.keys():
            return self.data[index]

        if index not in self.keys():
	    return

	field = getattr(self.dataset, index)
	if (index,) == field.dimensions:
	    self.data[index] = ma.array(field[:])
	    self.data[index].attributes = field.attributes
	    return self.data[index]
	else:
	    # Not sure the best way to do it
	    # To be transparent I should get the data as an array
	    # But there is always the risk to bring eveything by mistake
	    self.data[index] = ma.masked_values(field[index][:],field.missing_value)
	    #self.data[index].attributes = field.attributes
	    print dir(field[index])


	return field

from UserDict import UserDict

class MAFromNC(object):
    """   
    """  
    def __init__(self, ref):
        self.ref = ref
        self.shape = self.ref.shape
        self.dtype = self.ref.dtype
        self.units = self.ref.units
        #self.copy = self[:].copy # Not sure if it will work
    def __getitem__(self, key):
        if hasattr(self.ref,'missing_value'):
            return ma.masked_values(self.ref[key], self.ref.missing_value)
        else:
            return ma.array(self.ref[key])
    #def __setitem__(self, key, value):
    #    if type(value) == numpy.ma.core.MaskedArray:
    #        #self.ref[key] = value.filled()
    #        tmp = value.data 
    #        tmp[value.mask] = self.ref._v_attrs.missing_value
    #        self.ref[key] = tmp
    #    else:
    #        self.ref[key] = value



import netCDF4
class Modelo_from_nc(dict):
    """
    """
    def __init__(self, ncfpath):
        """
        """
        super(Modelo_from_nc, self).__init__()
        self.dataset = netCDF4.Dataset(ncfpath)
        self.close = self.dataset.close
        self.data = {}
        self.loaddata()
    def loaddata(self):
        self._set_datetime()
        # Improve this. It should be in a conditional and load only when
        #   requested. For now, this will work
        for var in ['xu_ocean', 'yu_ocean', 'st_ocean', 'st_edges_ocean', 
          'xt_ocean', 'yt_ocean', 'sw_ocean', 'sw_edges_ocean', 'hu']:
            if var in self.dataset.variables.keys():
                self.data[var] = ma.array(self.dataset.variables[var][:])

    def _set_datetime(self):
        """
	    Should I really do it right the way, automatically?
        """
	if 'time' in self.dataset.variables:
	    from coards import from_udunits
	    self.data['datetime'] = ma.array([ from_udunits(v, self.dataset.variables['time'].units) for v in self.dataset.variables['time'][:] ])
	return
    def keys(self):
        keys = self.data.keys()
	for k in self.dataset.variables.keys():
	    if k not in keys:
	        keys.append(k)
        return keys
    def __getitem__(self,index):
        """
        """
        if index in self.data.keys():
            return self.data[index]
        else:
            self.data[index] = MAFromNC(self.dataset.variables[index])
            return self.data[index]
        field = self.dataset.variables[index]
        if (index,) == field.dimensions:
            self.data[index] = ma.array(field[:])
            self.data[index].units = field.units
            return self.data[index]
        else:
            # Need to be able to get only a slice before convert 
            #   to masked array, otherwise is a problem for huge
            #   files
            data = ma.masked_values(field[:], field.missing_value)
            data.units = field.units
            data.long_name = field.long_name
            data.dimensions = field.dimensions
            return data
        #    return self.data[index]
        return
    def get_data(self,index,slice):
        """
        """
        field = self.dataset.variables[index]
        return ma.masked_values(field[slice], field.missing_value)


class Model_profiles(Modelo_from_nc):
    """


    """
    def __init__(self, ncfpath):
        super(Model_profiles, self).__init__(ncfpath)
        self.data = Modelo_from_nc(ncfpath)
        self.keys = self.data.keys
        # Unfinished
        return

    def nearest_profile(self,t,lon,lat,var):
        """
        """
        # --- Model ----
        # Define the closest time instant
        #dt = self.['datetime']-t
        #nt = numpy.arange(dt.shape[0])[dt == min(dt)][0]
        nt = numpy.absolute(ma.array([d.toordinal() for d in self['datetime']])-t.toordinal()).argmin()
        # Define the nearest point
        nx = numpy.absolute(self['xt_ocean']-lon).argmin()
        ny = numpy.absolute(self['yt_ocean']-lat).argmin()
        # Extract the temperature profile from the model
        profile = ma.masked_values(self.dataset.variables[var][nt,:,ny,nx],value=self.dataset.variables[var].missing_value)
        # Restrict the profile to deeper than 10m, and only good data
        # The find_z20 can't handle masked arrays
        #ind = (temp_model.mask==False) & (numpy.absolute(self['depth'])>10)
        #z = model_ref['depth'][ind]
        #t = temp_model[ind]
        return {'depth':self['st_ocean'],var:profile,'lon':self['xt_ocean'][nx],'lat':self['yt_ocean'][ny]}
        #return profile

    def around_profile(self, t, lon, lat, var, rmax):
        """
        
            !!ATENTION!! There are a lot of problems on this approach for big distances,
              including the borders and the poles.

            Improve it to actually calculate the distances.
        """
        nt = numpy.absolute(ma.array([d.toordinal() for d in self['datetime']])-t.toordinal()).argmin()
        #from fluid.common.distance import distance
        #Lon, Lat = numpy.meshgrid(model_ref['xt_ocean'], model_ref['yt_ocean'])
        ##L = distance(lon = Lon, lat = Lat, lon_c = lon, lat_c = lat)
        #fac = numpy.cos((lat+Lat)/2.*numpy.pi/180)
        #L = ((Lat-lat)**2+((Lon-lon)*fac)**2)**.5
        #L = L*60*1852

        nX = numpy.arange(self['xt_ocean'].shape[0])[abs(self['xt_ocean']-lon)<(rmax/1856./60.)]
        nY = numpy.arange(self['yt_ocean'].shape[0])[abs(self['yt_ocean']-lat)<(rmax/1856./60.)]

        profile = {'depth':self['st_ocean']}
        profile['lon'] = self['xt_ocean'][nX[0]:nX[-1]+1]
        profile['lat'] = self['yt_ocean'][nY[0]:nY[-1]+1]

        profile[var] = ma.masked_values(self.dataset.variables[var][nt,:,nY[0]:nY[-1]+1,nX[0]:nX[-1]+1],value=self.dataset.variables[var].missing_value)
        #profile = ma.masked_values(self.dataset[var][nt,:,nY:nY+1,nX:nX+1],value=self.dataset[var].missing_value)

        return profile




class Modelo_from_nca(UserDict):
    """
    """
    def __init__(self,ncfpath):
        """
        """
        self.data = {}
        self.dataset = []
        for ncf in ncfpath:
            self.dataset.append(Modelo_from_nc(ncf))
        #
        self._build_time()
        return
    def _build_time(self):
        time = numpy.array([])
        mask = numpy.array([], dtype='bool')
	for d in self.dataset:
            time = numpy.append(time, d['time'].data)
            mask = numpy.append(mask, ma.getmaskarray(d['time']))
	self.data['time'] = ma.masked_array(time,mask)
    def keys(self):
        keys = self.data.keys()
	for k in self.dataset[0].keys():
	    if k not in keys:
	        keys.append(k)
        return keys
    def __getitem__(self,index):
        """
        """
        if index in self.data.keys():
            return self.data[index]
        if index not in self.keys():
            return

        dims = self.dataset[0].dataset.variables[index].dimensions
        if 'time' in dims:
            #if 'time' == dims[0]:
            data = numpy.array([])
            mask = numpy.array([], dtype='bool')
            for d in self.dataset:
                time = numpy.append(data, d[index].data)
                mask = numpy.append(mask, ma.getmaskarray(d[index]))
            return ma.masked_array(time,mask)
        else:
	    return self.dataset[0][index]


#ncfpath = ["/stornext/grupos/ocean/simulations/exp030/dataout/ic200701/ocean/CGCM/20070101.ocean_transport.nc", "/stornext/grupos/ocean/simulations/exp030/dataout/ic200701/ocean/CGCM/20080101.ocean_transport.nc"]
#reload(mom4)
#x = Modelo_from_nca(ncfpath)
#x.keys()
#x['xu_ocean']
#x['ty_trans']

##if __name__ == '__main__':
#urlpath="http://opendap.ccst.inpe.br/Models/CGCM2.1/exp030/20070101.ocean_transport.nc"
#from pydap.client import open_url
#dataset = open_url(urlpath)
#import mom4
#reload(mom4)
#x = mom4.Modelo_from_dap(urlpath = urlpath)
#print dir(x)




import time
import sys
import math
import os
import re


def spin(times, interval=1):
    SPIN = '|\\-/'
    i = 0
    while i <= times:
        sys.stdout.write("\r%s" % SPIN[i % len(SPIN)])
        sys.stdout.flush()
        time.sleep(interval)
        i += 1


def layout(n, factor=1):

    def divisors(num):
        result = []
        step = 1
        while step <= math.ceil(num / 2):
            if num % step == 0:
                result.append(step)
            step += 1
        result.append(num)
        return result

    possibilities = [(n / d, d) for d in divisors(n) if n / d >= d]
    possibilities.sort(key=lambda x: abs(float(n / x[1] ** 2) - factor))
    return possibilities[0]


def get_output_files(cfg):
    """

        Example
        cfg = {'data_path': '/Tupa/simulations/exp048/dataout',
               'filename_pattern': 'cgcm2.2_currents_.*\.nc'}
    """
    # Create files list
    filenames = [f for f in os.listdir(cfg['data_path']) if re.match(cfg['filename_pattern'],f)]
    filenames.sort()
    # Include the path to the files
    ncfpaths = [os.path.join(cfg['data_path'],f) for f in filenames]
    return ncfpaths



#!/usr/bin/env python
# -*- coding: Latin-1 -*-
# vim: tabstop=4 shiftwidth=4 expandtab


import yaml
import re


def nml_decode(text):
    """
    """
    output = []
    text = re.sub("\t", " ", text)

    # dummy way to deal with comments. Need to improve it
    text = re.sub("\s*!.*\n", "\n", text)

    for namelists in re.findall("&((?:.*\n)+?)\s*/", text):
        g = re.match("\s*(?P<groupname>\w+)\n((?:.*\n)*)", namelists).groups()
        output.append("%s:\n" % g[0])
        #for p in re.findall("\s+(\w+)\s*=\s*(\.?.*\.?)\s*,?\s*(!.*)\n", g[1]):
        for p in re.findall("\s*(\w+)\s*=\s*(\.?.*\.?)\s*,?\s*\n", g[1]):
            tmp = re.sub(",$|\s*", "", p[1])
            if tmp == ".false." or tmp == '.FALSE.':
                tmp = False
            if tmp == ".true." or tmp == '.TRUE.':
                tmp = True
            output.append("  %s: %s\n" % (p[0], tmp))
    textout = "".join(output)
    return yaml.safe_load(textout)

# Working on
#group_pattern = """
#(&\s*
#(?P<groupname>\w+).*\n
#(.*\n)+
#)
#"""
#re.search(group_pattern, text, re.VERBOSE).groups()


def yaml2nml(cfg, key_order=None):
    """
    """
    output = []
    keys = cfg.keys()

    if key_order:
        key_order = list(key_order)
        if set(key_order) <= set(keys):
            key_order += (set(keys) - set(key_order))
        else:
            raise KeyError
    else:
        key_order = sorted(keys)

    for k in key_order:
        output.append(" &%s\n" % k)
        if cfg[k] is not None:
            parameters = cfg[k].keys()
            parameters.sort()
            for kk in parameters:
                v = cfg[k][kk]
                if v is False:
                    v = '.FALSE.'
                elif v is True:
                    v = '.TRUE.'
                else:
                    try:
                        if abs(int(v) - float(v)) == 0:
                            v = int(v)
                        else:
                            v = float(v)
                    except ValueError:
                        try:
                            v = float(v)
                        except ValueError:
                            try:
                                v.split(',')[1]
                            except ValueError:
                                v = "'%s'" % v
                            except IndexError:
                                v = "'%s'" % v
                            else:
                                v = "%s" % v
                output.append("    %s = %s,\n" % (kk, v))
        output.append("/\n\n")
    return "".join(output)


# Under work.
#filename = "/Users/castelao/work/projects/INPE/repos/Modelos/MOM4p1/src/mom4p1/ocean_diag/ocean_drifters.F90"
#filename = "/Users/castelao/work/projects/INPE/repos/Modelos/MOM4p1/src/mom4p1/ocean_core/ocean_barotropic.F90"
#f = open(filename)
#text = f.read()
#f.close()

charref = re.compile(r"""
    namelist\ +/(?P<namelist>\w*)/   # Name of the namelist
    (?P<parameters>
      (?:
        (?:           # Along one line
          \t?\ *
          (\w+)       # The Parameter
          ,?          # Comma and space if there is more than one
        )+
        \ *\&\ *
        \s
      )+
    )
""", re.VERBOSE)

#charref = re.compile(r'''namelist\ /(?P<namelist>\w*)/\ (?P<parameters>.*?)[^\&\ *]$''', re.DOTALL)
#parameters = map(strip, parameters)
#parameters = parameters.replace('&', '').split(',')
#parameters = map(strip, parameters)

charref = re.compile(r"""
        .*\n\t.*
        """, re.VERBOSE)


#charref.search(text).groupdict()
